/********************************************************************
  BasicDSP
  Copyright 2006-2007 
  Pieter-Tjerk de Boer PA3FWM pa3fwm@amsat.org
  Niels Moseley PE1OIT n.a.moseley@alumnus.utwente.nl
  License: GPLv2
********************************************************************/

#include "vu_meter.h"
#include "math.h"

#include <wx/dcbuffer.h>

BEGIN_EVENT_TABLE(wxVUMeter, wxControl)
	EVT_PAINT(wxVUMeter::OnPaint)
	EVT_ERASE_BACKGROUND(wxVUMeter::OnEraseBackground)
        EVT_SIZE(wxVUMeter::OnSize)
END_EVENT_TABLE()

IMPLEMENT_DYNAMIC_CLASS(wxVUMeter, wxControl)

wxPen *myYELLOW_PEN = new wxPen (wxColour(255, 255, 0), 1, wxSOLID);
wxBrush *myYELLOW_BRUSH = new wxBrush (wxColour(255, 255, 0), wxSOLID);

wxVUMeter::~wxVUMeter()
{
  dc_background.SelectObject(wxNullBitmap);
  delete background;
  delete tempbitmap;
}

bool wxVUMeter::Create(wxWindow *parent, wxWindowID id, const wxString &label,
	const wxPoint& pos,
	const wxSize& size)
{
	if (!wxControl::Create(parent, id, pos, size, wxSIMPLE_BORDER))
		return false;

	SetLabel(label);

	return true;
}


void wxVUMeter::OnSize(wxSizeEvent& event)
{
   if (background) 
   {
      dc_background.SelectObject(wxNullBitmap);
      delete background;
   }
   if (tempbitmap)
     delete tempbitmap;

   int cwidth  = GetClientRect().GetWidth();
   int cheight = GetClientRect().GetHeight();
   tempbitmap = new wxBitmap(cwidth, cheight, -1);
   background = new wxBitmap(cwidth, cheight, -1);

   dc_background.SelectObject(*background);
   dc_background.SetBrush(*wxWHITE_BRUSH);
   dc_background.SetPen(*wxWHITE_PEN);
   dc_background.DrawRectangle(GetClientRect());
   dc_background.SetFont(*wxSMALL_FONT);
   dc_background.SetPen(*wxBLACK_PEN);

   // draw scale
   wxCoord twidth, theight;
   wxCoord tdescent, tLeading;
   wxString stringbuf;
   for(int i=-48; i<=0; i+=6) {
      int xpos = db2pix(i);
      dc_background.DrawLine(xpos, 0, xpos, 5);
      stringbuf.Printf(_T("%d"),i);
      GetTextExtent(stringbuf, &twidth, &theight, &tdescent, &tLeading, wxSMALL_FONT);
      dc_background.DrawText(stringbuf, xpos - (twidth/2), 7);
   }

   // draw meter name
   GetTextExtent(GetLabel(), &twidth, &theight, &tdescent, &tLeading, wxSMALL_FONT);
   wxSize mysize = GetClientRect().GetSize();
   dc_background.DrawText(GetLabel(), 10, mysize.GetHeight()/2 + (-theight + tdescent) /2);
}




void wxVUMeter::SetLevel(float l)
{
  // clip the peaklevel to sane values
	if (l<1e-40) 
    level = 1e-40;
	else 
    level = l;

  // calculate the new peak level
  if (peaklevel < level || --peakcounter<=0) {
    peaklevel = level;
    peakcounter = 25;
  }

	Refresh();
}


int wxVUMeter::db2pix(float db)
{
	return static_cast<int>((db - dbmin) / (dbmax-dbmin) * GetClientRect().GetSize().GetWidth());
} 


void wxVUMeter::OnPaint(wxPaintEvent& event)
{
  wxPaintDC realdc(this);
  wxMemoryDC dc(&realdc);
  dc.SelectObject(*tempbitmap);
	// if we have double buffering, we should first draw the background color
	// ourselves. This is kind of a hack, but I don't know a better way to do this.
	// N.
	wxWindow *p = wxVUMeter::GetParent();
	if (p!=NULL)
	{	
		wxColour bk_color = p->GetBackgroundColour();
		wxBrush  bk_brush(bk_color);
		wxPen    bk_pen(bk_color);
		dc.SetBrush(bk_brush);
		dc.SetPen(bk_pen);
		wxRect rect = GetClientRect();
		dc.DrawRectangle(rect.x, rect.y, rect.x+rect.width, rect.y + rect.height);
		dc.SetBrush(*wxBLACK_BRUSH);
		dc.SetPen(*wxWHITE_PEN);
	};

	wxRect rect = GetClientRect();

	// plot 0..1 using logarithmic scale
	// db is in decibels
	float db = 20.0*log10(level);
	if (db<dbmin) db=dbmin;
	if (db>dbmax) db=dbmax;

  // calculate the peak position in pixels
  float dbpeak = 20.0*log10(peaklevel);
	if (dbpeak<dbmin) dbpeak=dbmin;
	if (dbpeak>dbmax) dbpeak=dbmax;
  int peakpos = db2pix(dbpeak);

	// calculate the number of pixels
	wxSize mysize = rect.GetSize();
	int pixels = db2pix(db);

	if (db <= -3)
	{
		// -3 dB or less
		dc.SetBrush(*wxGREEN_BRUSH);
		dc.SetPen(*wxGREEN_PEN);
		dc.DrawRectangle(0, 0, pixels, rect.GetHeight());
	}
	else
	{
		// over -3 dB
		dc.SetBrush(*wxGREEN_BRUSH);
		dc.SetPen(*wxGREEN_PEN);
		int c = db2pix(-3);
		dc.DrawRectangle(0, 0, c, rect.GetHeight());

		if (db <= 0)
		{
			dc.SetBrush(*myYELLOW_BRUSH);
			dc.SetPen(*myYELLOW_PEN);
			dc.DrawRectangle(c, 0, pixels-c, rect.GetHeight());
		}
		else
		{
			dc.SetBrush(*myYELLOW_BRUSH);
			dc.SetPen(*myYELLOW_PEN);
			int d = db2pix(0);
			dc.DrawRectangle(c, 0, d-c, rect.GetHeight());
			dc.SetBrush(*wxRED_BRUSH);
			dc.SetPen(*wxRED_PEN);
			dc.DrawRectangle(d, 0, pixels-d, rect.GetHeight());
		}
	}

  // draw the peak level
  dc.SetPen(*wxBLACK_PEN);

  dc.DrawLine(peakpos, 0, peakpos, rect.GetHeight());
  dc.DrawLine(peakpos+1, 0, peakpos+1, rect.GetHeight());

  //dc_background.SelectObject(*background);
  dc.Blit(0,0, rect.GetWidth(), rect.GetHeight(), &dc_background, 0, 0, wxAND, true);
  
  // blit the dc to the realdc!
  dc.SelectObject(wxNullBitmap);
  realdc.DrawBitmap(*tempbitmap, 0, 0);
}

wxSize wxVUMeter::DoGetBestSize() const
{
	wxCoord twidth, theight;
	wxCoord tdescent, tLeading;	

	GetTextExtent(_T("1234567890ABCDEFGHIJKLMNOPQRSTUVWXYZ"), &twidth, &theight, &tdescent, &tLeading, wxSMALL_FONT);
	return wxSize(GetRect().GetWidth(), theight + 9);
}

void wxVUMeter::OnEraseBackground(wxEraseEvent& event)
{
	#if wxCHECK_VERSION(2, 6, 3)
	// do nothing -- double buffering doesn't need an erase!
	#else
	// call standard erase function -- we don't have double buffering
	ClearBackground();	
	#endif
}

