/********************************************************************

  Wav streamer rev3.
  Code by N.A. Moseley 
  Copyright 2006-2007

  Note: only stereo files using PCM or IEEE floats are supported!
  supports 16,24,32 bit PCM and 32 bit float.
  stereo only!!!
  License: GPLv2

  rev3: first working version
  rev4: converted to unicode filenames & wxwidgets FileStream
********************************************************************/

#include "wav_streamer.h"

#define TEMPBUFFERSIZE 65536

WavStreamer::WavStreamer() : mutex(wxMUTEX_DEFAULT), wavefile(NULL), sample_index(0), temp_buffer(NULL)
{
}

WavStreamer::~WavStreamer()
{
  if (wavefile!=NULL) 
    delete wavefile;  // destructor automatically closes the file.
  if (temp_buffer != NULL)
    delete[] static_cast<char*>(temp_buffer);
}

int WavStreamer::OpenFile(const wxString &filename)
{
  wxMutexLocker locker(mutex);  // lock the mutex, unlocks automatically upon exit.

  // try to open the file
  wavefile = new wxFileInputStream(filename);
  if (wavefile == NULL)
    return -2;

  // check for RIFF file & size
  size_t bytes = wavefile->Read(chunk_type, 4).LastRead();
  if ((bytes != 4) || (strncmp(chunk_type, "RIFF", 4)!=0))
  {
    delete wavefile;
    wavefile = NULL;
    return -1;
  }
  bytes = wavefile->Read(&chunk_size, 4).LastRead();
  //bytes = fread(&chunk_size, 1, 4, wavefile);
  if (bytes != 4) 
  {
    delete wavefile;
    wavefile = NULL;
    return -1;
  }
  bytes = wavefile->Read(&chunk_type, 4).LastRead();
  //bytes = fread(&chunk_type, 1, 4, wavefile);   // not a chunk, but some other 4-letter type = 'WAVE'.
  if ((bytes != 4) && (strncmp(chunk_type, "WAVE", 4)!=0))
  {
    delete wavefile;
    wavefile = NULL;
    return -1;
  }

  riff_size = chunk_size + 8; // size of RIFF chunk + 4-byte ID + 4-byte chunksize.
  if (!FindChunk("fmt "))
  {
    // error, format chunk not found!
    delete wavefile;
    wavefile = NULL;
    return -1;
  }

  // now, read the format chunk
  bytes = wavefile->Read(&myformat, sizeof(WavFormatChunk)).LastRead();
  //bytes = fread(&myformat, 1, sizeof(WavFormatChunk), wavefile);
  if (myformat.wFormatTag == -2)  // WAVE_FORMAT_EXTENSIBLE case.. 
  {
    myformat.wFormatTag = 1;  // treat as PCM data, which should be the same when we have only 2 channels.
  }

  if (((myformat.wFormatTag != 3) && (myformat.wFormatTag != 1)) || (myformat.wChannels != 2))
  {
    delete wavefile;
    wavefile = NULL;
    return -1;  // wrong format!
  }
  // formats smaller than 16 bits are not supported!
  if (myformat.wBitsPerSample < 16)
  {
    delete wavefile;
    wavefile = NULL;
    return -1;
  }

  // if there are additional bytes in the format chunk, skip them.
  // important in the WAVE_FORMAT_EXTENSIBLE case!
  if (sizeof(WavFormatChunk) != chunk_size)
  {
    wavefile->SeekI(chunk_size - sizeof(WavFormatChunk), wxFromCurrent);
    //fseek(wavefile, , SEEK_CUR);
  }
  // now, search for the audio data and set the file offset pointers
  if (!FindChunk("data"))
  {
    delete wavefile;
    wavefile = NULL;
    return -1;
  }
  playback_start  = wavefile->TellI();
  playback_offset = playback_start;
  playback_end    = playback_start + chunk_size;

  // allocate the correct temporary buffer.
  if (temp_buffer!=NULL) delete[] static_cast<char*>(temp_buffer);
  temp_buffer = static_cast<void*>(new char[myformat.wBitsPerSample*TEMPBUFFERSIZE*2/8]);

  // don't forget to actually read the data
  ReadRawData(TEMPBUFFERSIZE);
  sample_index = 0;

  my_filename = filename;
  return 0;
}

bool WavStreamer::FindChunk(const char ID[4])
{
  long bytes_read = wavefile->TellI(); //ftell(wavefile);
  chunk_size = 0;
  // iterate until we get a format chunk..
  while ((riff_size > bytes_read) && (strncmp(chunk_type, ID, 4)!=0))
  {
    // skip the size of the chunk (except for the first RIFF chunk!)
    wavefile->SeekI(chunk_size, wxFromCurrent);
    //fseek(wavefile, chunk_size, SEEK_CUR);
    bytes_read += chunk_size;

    // read type of next chunk
    size_t bytes = wavefile->Read(chunk_type, 4).LastRead();
    //size_t bytes = fread(chunk_type, 1, 4, wavefile);
    bytes_read += bytes;
    bytes = wavefile->Read(&chunk_size, 4).LastRead();
    //bytes = fread(&chunk_size, 1, 4, wavefile);
    bytes_read += bytes;
  }
  if (riff_size < bytes_read)
    return false;

  return true;
}

int  WavStreamer::ReadRawData(int RequestedSamples)
{
  int bytes_per_sample;
  switch(myformat.wBitsPerSample)
  {
  case 16:
    bytes_per_sample = 2;
    break;
  case 24:
    bytes_per_sample = 3;
    break;
  case 32:
    bytes_per_sample = 4;
    break;
  default:
    return -1;  // error, wrong format!
  }
  int bytes_to_read = RequestedSamples * 2 * bytes_per_sample;  // RequestedSamples is in 'stereo', so 2 real samples.
  int offset = 0;
  while(bytes_to_read>0)
  {
    if (bytes_to_read > (playback_end - playback_offset))  // check for wrap-around
    {
      int read_amount = (playback_end - playback_offset);
      wavefile->Read(static_cast<char*>(temp_buffer) + offset, read_amount);
      //fread(static_cast<char*>(temp_buffer) + offset, 1, read_amount, wavefile);       //FIXME: handle errors
      wavefile->SeekI(playback_start);
      //fseek(wavefile, playback_start, SEEK_SET);
      offset += read_amount;
      playback_offset = playback_start;
      bytes_to_read -= read_amount;
    }
    else 
    {
      wavefile->Read(static_cast<char*>(temp_buffer) + offset, bytes_to_read);
      //fread(static_cast<char*>(temp_buffer) + offset, 1, bytes_to_read, wavefile);          //FIXME: handle errors
      playback_offset += bytes_to_read;
      bytes_to_read = 0;
    }    
  }
  sample_index = 0;
  return RequestedSamples;
}

void WavStreamer::FillBuffer(float *stereo_buffer, int stereo_samples)
{
  wxMutexLocker locker(mutex);  // lock the mutex, unlocks automatically upon exit.
  if (wavefile==NULL)
  {
    // this is compatible with the IEEE 756 floating-point format!
    memset(stereo_buffer, 0, sizeof(float)*2*stereo_samples);
    return;
  }
  else
  {
    int mono_samples = stereo_samples * 2;
    if (myformat.wFormatTag == 1) // PCM
    {
      switch(myformat.wBitsPerSample)
      {
      case 16:
        {
          short *ptr = static_cast<short*>(temp_buffer);
          for(int i=0; i<mono_samples; i++)
          {
            if (sample_index > (TEMPBUFFERSIZE*2-1)) ReadRawData(TEMPBUFFERSIZE);
            stereo_buffer[i] = static_cast<float>(ptr[sample_index]) / 32768.0f;
            sample_index++;
          }
        }
        break;
      case 24:
        {
          signed char *ptr = static_cast<signed char*>(temp_buffer);
          long v = 0;
          int index = sample_index * 3;
          for(int i=0; i<mono_samples; i++)
          {
            if (sample_index > (TEMPBUFFERSIZE*2-1)) 
            {
              ReadRawData(TEMPBUFFERSIZE);
              index = 0;
            }
            v = static_cast<int>(ptr[index]) << 8;
            v |= static_cast<int>(ptr[index+1]) << 16;
            v |= static_cast<int>(ptr[index+2]) << 24;
            index += 3;
            stereo_buffer[i] = static_cast<float>(v) / 2147483648.0f;
            sample_index++;
          }
        }
        break;
      case 32:
        {
          int *ptr = static_cast<int*>(temp_buffer);
          for(int i=0; i<mono_samples; i++)
          {
            if (sample_index > (TEMPBUFFERSIZE*2-1)) ReadRawData(TEMPBUFFERSIZE);
            stereo_buffer[i] = static_cast<float>(ptr[sample_index]) / 2147483648.0f;
            sample_index++;
          }
        }
        break;
      }
    }
    else if (myformat.wFormatTag == 3)
    {
      // uncompressed float data
      float *ptr = static_cast<float*>(temp_buffer);
      for(int i=0; i<mono_samples; i++)
      {
        if (sample_index > (TEMPBUFFERSIZE*2-1)) ReadRawData(TEMPBUFFERSIZE);
        stereo_buffer[i] = ptr[sample_index]; // FIXME: scaling??
        sample_index++;
      }
    }
  }
}

bool WavStreamer::GetFormat(WavFormatChunk &output) const
{
  if (wavefile == NULL) return false;
  output = myformat;
  return true;
}
